.nr LL 5.5i
.TL
Unix Portability Suite
.AU
Brian Campbell
.AI
Quantum Software Systems
.AB
.AE

.SH
INTRODUCTION
.PP
QNX 4.0 is intended to be a POSIX 1003.1 conforming (pending
verification) operating system.  This means it supports a standard set
of system calls, intrinsics and utilities and that POSIX conforming
programs moved from other conforming operating systems should be
readily compilable and executable under QNX 4.0 (assuming no operating
system or hardware specific assumptions were made).
.PP
Unix is a generic term (AT&T must legally disagree here) describing a
number of quite different implementations of an operating system.  The
two most common variants are AT&T's SYSV and the University of
California at Berkley's BSD.  Another variant which deserves mention
is Microsoft's XENIX, which is the most common Intel based version.
POSIX, for the most part, resembles SYSV but borrows some features
from BSD as well.
.PP
Since POSIX functionality is largely composed of features from
existing long-lived unix variants, most host-specific routines are
readily implementable on top of the required POSIX functions.  Other
than furnishing some cover functions which provide duplicate
functionality the process or porting unix applications is simple.
.PP
There is a plethora of C software available, in source form, for
unix operating systems.  This includes editors, spreadsheets, database
management systems, varied interpreter and compilers, communication
programs, mail and news readers and games.
.PP
The goal of the portability library is to provide as many of the
capabilities of both SYSV and BSD as possible to make the task of
porting applications as easy as possible.
.SH
MAKE
.PP
The make utility is a useful development tool.  It controls
recompilation of specific sources files by the declaring what other
files that source file is dependent on.  For example, all sources in a
project often include a project specific header file.  If that header
file is changed it may affect any or all of the project source files.
The make utility compares the times of all the source files with their
dependents and determines which need to be recompiled.
.PP
The makefile (the file where the source file dependencies are
described) also declares what flags and options are used in the
compile and link steps.  The most important of these options is the
which symbols will be defined for the compile stage.
.SH
DEFINES
.PP
The define preprocessor directive is used by C compilers to declare
constants and macros that will be used and re-used in program source.
They are also used to control, via conditionals, which portions of a
program will be seen by the compiler.  Manifest defines are used to
bridge portability gaps across hardware platforms and operating
systems.  For example, the same functionality is often possible under
DOS and QNX, but the method is often radically different because of
differently named operating system support headers or routines.  To
support both systems with a single source file requires that the DOS
and QNX specific portions be separated from each other.  This is
accomplished with the C preprocessors conditional statements if,
ifdef, else and endif.
.PP
The most common symbols used to control program compilation in unix
environments are SYSV, BSD, USG, V7 and XENIX.  There are several
variations on these as well including SVR3, SVR4 (for System 5
releases three and four) and BSD42, BSD43 (for Berkley's releases 4.2
and 4.3) etc.  A new symbol is becoming more common, _POSIX_SOURCE
which identifies source which make use of POSIX-specified libraries.
.PP
Since application creation is generally determined by makefiles, this
is usually a good spot to search for preprocessor definitions which
will be passed to the compiler.  In general these sections are usually
well commented.  Often there will be several sets of pre-built sets of
compiler flags setup for common environments commented out.  Most of
the time the SYSV variant will work best with QNX.
.PP
A quick glance at the dependencies in the makefile may also point to a
common header file.  There will often be a header file named config.h,
common.h or sysdep.h which will have several feature controlling
defines listed, which may be individually commented out or undefined
to specify the behaviour of the resulting program.
.SH
HEADERS
.PP
Manifest defines may also be used to control the inclusion of system
header files.  A parallel set of routines may exist in a particular
header file for one set of systems, while another system will define
the routines and their constants in another header.  For example, BSD
declares their string routines in <strings.h> while SYSV references
<string.h>.  Another common case is to include <memory.h> on other
systems where <string.h> should be included under QNX.  It might be
advantageous to ln <string.h> <memory.h>
.PP
Header files appearing under the sys subdirectory are by definition
system dependent.  There are a number of names which are always
available (such as <sys/stat.h> and <sys/types.h>) which define system
specific structure layouts and data-types.  Other headers describe
system specific features not available (or available under other
names) in other operating systems.
.SH
LIBRARY ARCHIVES
.PP
Unix libraries are collections of files, usually compiler/assembler
generated object files, produced by the ar utility.  The ar utility
records the filename, owner, modification time, permissions and size.
Archive files generated by the ar utility, by convention, end with .a.
Since the ar utility is a general purpose utility for packaging
text or binary files into a single file, another utility, ranlib, is
required to let Unix linkers determine what names are declared in the
contained object files.  The ranlib utility commonly adds a __.SYMDEF
file which simply contains a list of exported functions.
.PP
Unix development systems often make use of object archive files for
moving in several object files to avoid clutter.  Since the ar utility
records the name and timestamp of the contained objects, the Unix make
utility can extract those timestamps to determine if the named object
file needs recompilation.  Indeed, there is a special syntax used in
Unix makefiles to specify object files contained in archive files.
.PP
QNX libraries closely parallel DOS libraries.  There is a special
utility, wlib, which is used to add, delete or replace objects files
in the library file.  QNX, and DOS, library files end with .lib.
Since QNX and DOS library files are compatible, DOS libraries (and
object files for that matter) may be used in linking QNX executables,
and QNX libraries and object files may be used in linking DOS
executables.  QNX libraries contain only the name of the contained
object files.
.PP
Since QNX libraries do not record the date of the contained files only
the timestamp of the library itself can be used by the make utility to
determine if a contained object file needs recompilation.  This can
result in unnecessary compiles.
.SH
COMPILING
.PP
There are some useful changes which should be made to the makefile (or
possibly in the configure script) which affect your compilation
environment.
.PP
Unless you know before-hard what model you will compiling in and are
sure that you won't change your mind later, it helps to declare the
following to macros:
.QP
	M=c
	CC=cc -m$M -g1
	CFLAGS=-j -Osal -v -w3 -zc -Wc,-fi=unix.h -Wc,-s -zt512
.PP
This says your default compilation model is the compact model (c) and
that the C compiler should be invoked with the -mc option (the M macro
is expanded) to specify compiling (and linking) in the compact model.
.PP
The -Wc,-fi=unix option tells cc that it should pass the -fi=unix.h
option to the C compiler stage (wcc).  The -fi=unix forces an include
of the <unix.h> header file.  This file declares the prototypes for
the functions in the unix portability library.
.PP
The -g1 option includes line number information in the object and load
files so that if the program crashes some indication of where it went
awry is available.  Line number information takes little space in the
object and load files.  It is also possible to compile with -g2 (or -g
for short) to indicate full symbolic debugging information.  This adds
significantly to the size of the executable (although the debugging
information isn't loaded at execution time).
.PP
Compiling with -g2 can affect code optimization.  If you are testing
and modifying a program in an effort to get it to work, it is better
to include just the line number information (-g1) and compile with
whatever optimizations the final version will use.  The version
compiled and linked with -g1 can simply be stripped when finished and
the resulting executable will be identical to one compiled without any
debug options.
.PP
The -j flag indications that the char type should be signed unless
modified with the unsigned modifier.  Since the signed modifier is a
relatively new ANSI invention, some unix programs may make subtle
assumptions about have a signed character type.  This may also produce
better or smaller code.
.PP
The -Osal option indicates the optimizations to perform.  The s
portion indicates to optimize for space (t indicates speed).  The a
option indicates the compiler may relax its alias checking (the code
optimizer will assume that global variables are not indirectly
referenced through pointers).  This assumption may reduce the size of
code that is generated.
.PP
The -v option tells the compiler to generate a corresponding .def file
for each source file it processes.  The .def file contains prototype
definitions for all the functions defined in a source file.  If there
are any warnings about missing prototypes the .def file can be
consulted and the appropriate prototype moved into a header file which
will be included by the offending module.  Alternatively, if the
application has no prototypes whatsoever, the .def files can be merged
into a single file, often proto.h, which would be included by a
central header file or by each of the source files.  Both serve to
ensure that all functions are called with prototype information for
that function in scope.
.PP
The -w3 option tells the compiler to issue warnings of severity type
1, 2 and 3.  This is the highest warning level the compiler currently
supports (there is a -w4 which specifies -w3 and that the messages
should be logged to .err files).  The intent here is that any warnings
the compiler generates should be acted on; missing prototypes should
be added, type mismatches should be fixed or appropriately cast.
.PP
One option which should be avoided the program is thoroughly tested
under a variety of conditions is -Wc,-s.  This option disables stack
checking.  Stack checking is some code the compiler emits at the
beginning of every function to determine if the stack consumed by the
function will overflow the stack.  Unix applications generally assume
the stack exists in a large flat data space.  Under QNX the stack is
limited to a 64k segment which it must share with the default data
segment.
.PP
The -zt option comes in handy if the linker complains that the "size
of DGROUP exceeds 64k".  By default, compiling with a large data model
(compact and large) only moves objects into far segments (as opposed
to the default near segment) if the object is larger than 32k.  The
-zt option specifies a data-size threshold; objects larger than this
value should be moved into far segments.
.PP
Note also, the stack segment also lives in the near DGROUP segment.
If you run out of stack space then you'll either have to decrease the
stack requirements of the offending routine or make more space in the
DGROUP segment by specifying a data-size threshold value.
.SH
LINKING
.PP
Standard unix development systems' linking loader is named ld.
Normally the ld utility is invoked by specifying to the cc utility
that an executable is to be produced, usually with the -o option.
Under QNX the ld utility is named wlink and can be similarly be
invoked through the -o option of cc.  ld is usually only called
directly to modify the startup object files or libraries to be linked
with the application (cc does some magic).
.SH
PECULIARITIES
.NH
CONFIGURE SCRIPTS
.PP
It is common for public domain applications to come with a
configuration shell script.  The most prevalent of these is generally
named Configure and was originally designed by Larry Wall (also the
author of author of the patch utility which is used to apply a series
of source code modifications given some surrounding context).
.PP
This is a massive script, often 40-60k in size.  It makes use of many
common unix shell features and utilities to determine such things as
the operating system you're running, what library functions are
available and what unix utilities are available and where.  Some of
the utilities used (most notably ar, discussed later) are not present
in QNX and severely harm the usefulness of the script.  [ar is used to
scan the library to decide what functions are available for use].
.PP
Another configuration trick is for a makefile to pass itself (or a
subsidiary makefile) through the C preprocessor, making use of the
pre-defined macros and conditional statements to emit a system specific
makefile.  This works best with wcc -i=/usr/include -p (not cc -E).
.NH
MODELS/LIMITS
.PP
Intel 80286 machines are 16-bit machines with 20-bit address spaces.
As such, there are usually five (or six) compile models with this
architecture: 
.PP
There are two code models:
.RS
.IP o
The small code model is one in which all calls to functions are near
calls.  In a near call the destination address is a 16-bit offset
relative to the segment value in the CS segment register.  Hence, in a
small model, all code comprising your program, including library
functions, must be less than 64k.
.IP o
The big code model is one in which all calls to functions are made
with far calls.  In a far call the destination address is a 32-bit
value (a 16-bit segment value and a 16-bit offset value relative to
that segment).  This model allows the size of the code comprising your
program to exceed 64k.
.RE
.PP
There are three data models:
.RS
.IP o
The small data model is one in which all references to data are made
with near pointers.  Near pointers are 16-bit values which are offset
relative to the segment value in the DS segment register.  Hence, in a
small data model, all data comprising your program, including near
library data, must be less than 64k.
.IP o
The big data model is one in which all references to data are made
with far pointers.  Far pointers are 32-bit values (a 16-bit segment
value and a 16-bit offset relative to that segment).  This removes the
64k limitation on data size imposed by the small data model.  However,
when a far pointer is incremented, only the offset is adjusted.  The
compiler assumes that the offset portion of a far pointer will not be
incremented beyond 64k.  The compiler will assign an object to a new
segment if the grouping of data in a segment will cases to object to
cross a segment boundary.  Implicit in this is the requirement that no
individual object exceeds 64k bytes.  For example, an array of 40,000
integers does not fit into the big data model.  An object such as this
should be described as huge.
.IP o
The huge data model is one in which all references to data are made
with far pointers.  This is similar to the big data model.  However,
in the huge data model, incrementing a far pointer will adjust the
offset and the segment if necessary.  The limit on the size of an
object pointed to by a far pointer imposed by the big data model is
removed in the huge data model.
.RE
.PP
The unix compilation model is most similar to a large code and large
data model.  Code and data pointers are both 32-bits and objects are
not restricted to a maximum of 64k bytes in length.  The difference
between QNX large model and the flat unix model is that integers under
QNX are only 16-bits (although longs are 32-bits) while under unix
both integers and longs are 32-bits.  This last situation has led to
much unix software which assumes that ints are 32-bits and that
pointers may be stored in them, and converted to/from integers.  It
also has an effect on almost all functions since many functions take
integer parameters (such as read and write) and under unix the range
of integer is much larger than the range of an integer under QNX (2^16
times larger).
.PP
It is also necessary to watch out for, and harder to find, instances
where the compiler folds constants in an expression into the default
integer type where the value exceeds the range of an integer.  For
example, with 32-bit integers
.PP
	long offset = 75 * 500;
.LP
will assign the value 37,500 to the long variable offset, while with
16-bit integers the value assigned will be -28,306.
.PP
Some operating systems allow dereferencing the NULL pointer.  That is,
retrieving the value stored at offset 0.  Generally if the OS allows
this the value returned is zero (allowing sloppy code to walk one past
the end of a null-terminated list before realizing it hit a null
pointer or allowing the null-pointer to be passed to a string routine
which can blithely dereference the value and immediately see the
string's terminating NUL character.
.PP
With QNX, small data model programs may dereference the NULL pointer
with generating an exception, but the value returned is non-zero.  The
C library startup code fills in 32 bytes at offset zero to contains
the value 1.  When an small data program terminates, the exit code
checks to see if the 32 bytes at offset zero are all still one's and
if not issues the diagnostic:
.PP
	** NULL assignment detected.
.PP
This indicates that at some point your program modified the data at an
offset between 0 and 32.
.PP
Large data model programs which attempt to dereference NULL generate a
segmentation violation (signal SIGSEGV is raise'd on the program).
This can be a quick way to find out if and where a program attempts to
dereference or modify data at NULL (such as when an integer is assumed
to be the same size as a pointer and is passed to a routine which
expects a far pointer causing an invalid segment).
.NH
PROTOTYPES
.PP
It is important to included function prototypes (via standard ANSI or
POSIX header files where possible) for any routines called in an
application.  This is more important with QNX than other operating
systems because Watcom's optimizing compiler uses an extremely
efficient yet non-standard calling convention.  All parameters are
passed in registers until they are all used at which point the
remainder are pushed on the stack like in standard C compilers.
.PP
This makes the code highly efficient and often smaller since several
unnecessary pushes and pops are de-necessitated.  However, for the
compiler to assign arguments to registers it must know the sizes of
the parameters (for example, long integers or far pointers require two
16-bit registers).
.PP
If function receives or returns a type which is wider than an integer
a prototype is absolutely required.  Functions which take a variable
number of arguments also absolutely require a prototype (discussed
later).
.NH
SIGNALS/SETJMP
.PP
Old-style unix signals behave differently than POSIX signals.
Traditional unix signals are one-shot affairs.  When a signal is
generated the signal's action is reset to the default action and the
appropriate signal handler (if a program has set one up) is called.
The signal handler may return or longjmp() to a higher stack frame.
.PP
With POSIX compliant systems, when a signal is generated it is blocked
(a POSIX signal action which does not exist under older unices).  The
signal handler is called and if it returns (does not yet longjmp or
siglongjmp to a higher stack-frame) the signal is unblocked.  The
signal handler remains associated with the signal.  If the same signal
was generated during the execution of the signal handler it will be
unblocked when and if the signal handler returns and the process
repeats.
.PP
POSIX has specified an extended version of the setjmp/longjmp function
pair -- sigsetjmp/siglongjmp.  The latter pair add the ability to save
and restore the signal mask.  If longjmp() is used in a signal handler
it is important to use the unix compatible signals or to use
siglongjmp() in the signal handler instead (this requires setting up
the jump buffer with sigsetjmp at some earlier point).  Otherwise
longjmp'ing out of a signal handler (with POSIX signals) will leave
the signal masked, effectively disabling its further generation and
actioning.
.PP
With the unix compatibility library either behaviour is possible.  By
default, linking against the unix library gives the original old-style
unix behaviour.  [Signals are not blocked during the execution of the
signal handler.  The signal action is reset to the default immediately
before execution of the signal handler and must be re-vectored if
required]
.NH
TERMCAP/TERMINFO
.PP
Full-screen unix applications will generally make use of the curses
and/or termcap routines.  The termcap reference a terminal capability
database (/etc/termcap) which describes all the appropriate escape
sequences, including cursor positioning, screen clearing, attribute
changes, etc, for many terminals.  Termcap capabilities are always
described as two character sequences:  cl for clear screen, ho for
home cursor, etc.
.PP
A newer terminal capability database called terminfo exists and
references files under the /usr/lib/terminfo tree.  [The directory
immediately below is the first letter of the terminal name, and in the
directory below it are all the terminal description files for terminals
beginning with that letter].
.NH
VARIADIC FUNCTIONS
.PP
Functions which take a variable number of arguments are commonly used
for output formatting where there can be any number of arguments in the
format string.  The Watcom compiler has several different calling
sequences depending on the number and width of the parameters in a
function (enregistering as many as possible).  However, since a
variadic function's arguments are accessed through a macro all
parameters must be passed on the stack.  For this reason, it is very
important to include a prototype for any function which receive's a
variable number of arguments (otherwise it will try to pass the
arguments in registers).
.PP
Unix code often has the nasty habit of declaring variadic functions as
accepting several long parameters.  The type of these parameters is
often ignored in the function body; often the only action is to pass
each of these arguments (several of which may not have even been
passed in) to a lower-level variadic routine such as printf.
.PP
It is crucial for applications which make use of the low-level open()
routine to include <fcntl.h> for this function's prototype.  open() is
a function which accepts an optional third argument.  If this prototype
is not in scope the compiler will use the more efficient, but wrong
for this function calling convention quite likely making all file
opens fail.
.PP
QNX variadic functions must be declared in the ANSI-specified manner.
<stdarg.h> must be included in the file defining the function to use
the argument access macro, va_arg().
.PP
.nf
#include <stdarg.h>

void log_error(char level, ...) {
	if (level == 'E' || verbose) {
		va_list args;
		char *fmt;

		va_start(args, level);
		fmt = va_arg(args, char *);
		fprintf(stderr, "%s: '%c' ", progname, level);
		vfprintf(stderr, fmt, args);
		va_end(args);
	}
}
